/*
 * Copyright (c) 2024 Andrea Biscuola <a@abiscuola.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/stat.h>

#include <libgen.h>
#include <limits.h>
#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <curl/curl.h>
#include <api.h>

#include "compat.h"
#include "fchat.h"

#define TSIZE	32

static struct fchat		*makenew(const char *, const char *, const char *);
static uint64_t		 parseid(const char *);
static time_t		 parsetime(const char *);
static time_t		 parselocaltime(const char *);
static size_t		 parsesize(const char *);
static int			 login(struct fchat *);
static struct fchatmsg	*msgnew(struct fchatmsgs *);

int
fchaterror(struct fchatmsg *msg)
{
	return msg ? msg->error : -1;
}

char *
fchatfiledownload(struct fchat *hnd, struct fchatmsg *msg)
{
	char	lurl[FCHATMAXURLSIZE];
	char	*rb;

	if (snprintf(lurl, FCHATMAXURLSIZE, "%s/chat-download/%llud/%s",
	    hnd->url, msg->id, msg->fname) >= FCHATMAXURLSIZE)
		return NULL;

	if (apireset(hnd->capi, "GET", lurl) == -1)
		return NULL;

	if (apido(hnd->capi) == -1)
		return NULL;

	/* The size is provided by fchatfilesize */
	if (apidump(hnd->capi, &rb) == -1)
		return NULL;

	return rb;
}

char *
fchatfilemime(struct fchatmsg *msg)
{
	return msg ? msg->fmime : NULL;
}

char *
fchatfilename(struct fchatmsg *msg)
{
	return msg ? msg->fname : NULL;
}

size_t
fchatfilesize(struct fchatmsg *msg)
{
	return msg ? msg->fsize : 0;
}

void
fchatfree(struct fchat *hnd)
{
	if (!hnd)
		return;

	apifree(hnd->capi);
}

uint64_t
fchatid(struct fchatmsg *msg)
{
	return msg ? msg->id : 0;
}

int
fchatinit(void)
{
	return apiglobalinit(NULL);
}

int
fchatisdel(const struct fchatmsg *msg)
{
	return msg->mdel;
}

time_t
fchatlocaltime(struct fchatmsg *msg)
{
	return msg ? msg->lms : 0;
}

void
fchatmsgsfree(struct fchatmsgs *msgs)
{
	struct fchatmsg	*msg, *tmp;

	if (!msgs)
		return;

	STAILQ_FOREACH_SAFE(msg, msgs, next, tmp) {
		STAILQ_REMOVE(msgs, msg, fchatmsg, next);

		free(msg->text);
		free(msg);
	}

	free(msgs);
}

struct fchat *
fchatnew(const char *url, const char *u, const char *p)
{
	struct fchat	*new;

	if (!url || !u || !p)
		return NULL;

	new = makenew(url, u, p);
	if (!new)
		return NULL;

	if (login(new) == -1) {
		fchatfree(new);
		return NULL;
	}

	return new;
}

int
fchatnext(struct fchatmsgs *msgs, struct fchatmsg **msg)
{
	if (!msgs)
		return -1;

	if (!*msg)
		*msg = STAILQ_FIRST(msgs);
	else
		*msg = STAILQ_NEXT(*msg, next);

	return *msg ? 0 : -1;
}

struct fchatmsgs *
fchatpoll(struct fchat *hnd, int64_t num)
{
	char		lurl[FCHATMAXURLSIZE];
	struct fchatmsgs	*msgs;
	struct fchatmsg	*np;
	struct apiv	v;
	int		parse;

	if (!hnd)
		return NULL;

	if (snprintf(lurl, FCHATMAXURLSIZE, "%s/chat-poll?raw=1&name=%lld",
	    hnd->url, num) >= FCHATMAXURLSIZE)
		return NULL;

	if (apireset(hnd->capi, "GET", lurl) == -1)
		return NULL;

	switch (apido(hnd->capi)) {
	case 0:
		/* FALLTHROUGH */
	case -2:
		break;
	default:
		return NULL;
	}

	msgs = malloc(sizeof(struct fchatmsgs));
	if (!msgs)
		return NULL;

	STAILQ_INIT(msgs);

	parse = 0;
	np = NULL;
	while (apijsonnext(hnd->capi, &v) != -1) {
		switch (v.type) {
		case APIJSONVAL:
			break;
		case APIJSONARR:
			parse = !parse;
			break;
		case APIJSONOBJ:
			if (!parse)
				break;
			np = msgnew(msgs);
			if (!np) {
				fchatmsgsfree(msgs);
				return NULL;
			}
			break;
		case APIJSONSTR:
			if (!parse)
				break;
			if (strcasecmp(v.val, "isError") == 0) {
				np->error = 1;
			} else if (strcasecmp(v.val, "msgid") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				np->id = parseid(v.val);
				if (np->id == 0) {
					fchatmsgsfree(msgs);
					return NULL;
				}
			} else if (strcasecmp(v.val, "mdel") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				np->mdel = parseid(v.val);
				if (np->mdel == 0) {
					fchatmsgsfree(msgs);
					return NULL;
				}
			} else if (strcasecmp(v.val, "xfrom") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				if (strlcpy(np->user, v.val,
				    FCHATMUSRS) >= FCHATMUSRS) {
					fchatmsgsfree(msgs);
					return NULL;
				}
			} else if (strcasecmp(v.val, "mtime") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				np->ts = parsetime(v.val);
			} else if (strcasecmp(v.val, "lmtime") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				np->lms = parselocaltime(v.val);
			} else if (strcasecmp(v.val, "xmsg") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				np->text = strdup(v.val);
				if (!np->text) {
					fchatmsgsfree(msgs);
					return NULL;
				}
			} else if (strcasecmp(v.val, "fsize") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				np->fsize = parsesize(v.val);
			} else if (strcasecmp(v.val, "fname") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				if (strlcpy(np->fname, v.val,
				    PATH_MAX) >= PATH_MAX) {
					fchatmsgsfree(msgs);
					return NULL;
				}
			} else if (strcasecmp(v.val, "fmime") == 0) {
				if (apijsonnext(hnd->capi, &v) == -1) {
					fchatmsgsfree(msgs);
					return NULL;
				}
				if (strlcpy(np->fmime, v.val,
				    FCHATMIMEMAX) >= FCHATMIMEMAX) {
					fchatmsgsfree(msgs);
					return NULL;
				}
			}
		}
	}

	return msgs;
}

int
fchatsend(struct fchat *hnd, const char *text, char *file)
{
	char	lurl[FCHATMAXURLSIZE];
	struct stat	fs;
	char	*b;
	FILE	*f;
	struct tm	*td;
	char	ts[TSIZE];
	time_t	now;

	if (!hnd || !text)
		return -1;

	if (snprintf(lurl, FCHATMAXURLSIZE, "%s/chat-send", hnd->url)
	    >= FCHATMAXURLSIZE)
		return -1;

	if (apireset(hnd->capi, "POST", lurl) == -1)
		return -1;

	if (apiaddformfield(hnd->capi, "msg", text, NULL, 0) == -1)
		return -1;

	now = time(NULL);
	td = localtime(&now);
	if (strftime(ts, TSIZE, "%FT%T%z", td) == 0)
		return -1;
	if (apiaddformfield(hnd->capi, "lmtime", ts, NULL, 0) == -1)
		return -1;

	if (file) {
		if (stat(file, &fs) == -1)
			return -1;
		f = fopen(file, "r");
		if (!f)
			return -1;
		b = malloc(fs.st_size * sizeof(char));
		if (!b) {
			fclose(f);
			return -1;
		}
		if (fread(b, sizeof(char), fs.st_size, f) < (size_t)fs.st_size) {
			free(b);
			fclose(f);
			return -1;
		}
		if (apiaddformfield(hnd->capi, "file", b, basename(file),
		    fs.st_size) == -1) {
			free(b);
			fclose(f);
			return -1;
		}
	}

	if (apido(hnd->capi) == -1)
		return -1;

	return 0;
}

char *
fchattext(struct fchatmsg *msg)
{
	return msg ? msg->text : NULL;
}

time_t
fchattime(struct fchatmsg *msg)
{
	return msg ? msg->ts : 0;
}

char *
fchatuser(struct fchatmsg *msg)
{
	return msg ? msg->user : NULL;
}

static uint64_t
parseid(const char *str)
{
	return (uint64_t)strtonum(str, 0, INT64_MAX, NULL);
}

static time_t
parsetime(const char *str)
{
	struct tm	ts;

	if (!strptime(str, "%FT%T%z", &ts))
		return 0;

	 return timegm(&ts);
}

static time_t
parselocaltime(const char *str)
{
	struct tm	ts;

	if (!strptime(str, "%FT%T", &ts))
		return 0;

	 return timegm(&ts);
}

static size_t
parsesize(const char *str)
{
	return (size_t)strtonum(str, 0, INT64_MAX, NULL);
}

static struct fchat *
makenew(const char *url, const char *u, const char *p)
{
	struct fchat	*new;

	new = calloc(1, sizeof(struct fchat));
	if (!new)
		return NULL;

	if (strlcpy(new->url, url, FCHATMAXURLSIZE) >= FCHATMAXURLSIZE) {
		fchatfree(new);
		return NULL;
	}
	if (strlcpy(new->user, u, FCHATMUSRS) >= FCHATMUSRS) {
		fchatfree(new);
		return NULL;
	}
	if (strlcpy(new->pass, p, FCHATMAXPASSSIZE) >= FCHATMAXPASSSIZE) {
		fchatfree(new);
		return NULL;
	}

	return new;
}

int
login(struct fchat *hnd)
{
	char	lurl[FCHATMAXURLSIZE];

	if (strlcpy(lurl, hnd->url, FCHATMAXURLSIZE) >= FCHATMAXURLSIZE)
		return -1;
	if (strlcat(lurl, "/login", FCHATMAXURLSIZE) >= FCHATMAXURLSIZE)
		return -1;

	if (hnd->capi)
		apifree(hnd->capi);

	hnd->capi = apiinit("POST", lurl);
	if (!hnd->capi)
		return -1;

	if (apisetcookiefile(hnd->capi, NULL) == -1)
		return -1;
	if (apiaddformfield(hnd->capi, "u", hnd->user, NULL, 0) == -1)
		return -1;
	if (apiaddformfield(hnd->capi, "p", hnd->pass, NULL, 0) == -1)
		return -1;

	if (apido(hnd->capi) == -1)
		return -1;

	if (snprintf(lurl, FCHATMAXURLSIZE, "%s/chat?cli", hnd->url)
	    >= FCHATMAXURLSIZE)
		return -1;

	if (apiaddheader(hnd->capi, "Referer", lurl) == -1)
		return -1;

	return 0;
}

static struct fchatmsg *
msgnew(struct fchatmsgs *pr)
{
	struct fchatmsg	*new;

	new = calloc(1, sizeof(struct fchatmsg));
	if (!new)
		return NULL;

	STAILQ_INSERT_TAIL(pr, new, next);

	return new;
}
