/*
 * Copyright (c) 2024 Andrea Biscuola <a@abiscuola.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "compat.h"
#include "irc.h"

static struct ircmsg	*newmsg(void);
static int		 parsetags(char *, struct ircmsg *);

struct ircsrv *
ircnew(FILE *in, FILE *out)
{
	struct ircsrv	*srv;

	if (!in || !out)
		return NULL;

	srv = calloc(1, sizeof(struct ircsrv));
	if (!srv)
		return NULL;

	srv->in = in;
	srv->out = out;

	if (gethostname(srv->host, IRCHOSTLEN) == -1) {
		ircfree(srv);
		return NULL;
	}

	return srv;
}

void
ircfree(struct ircsrv *srv)
{
	if (!srv)
		return;

	if (srv->in)
		fclose(srv->in);

	if (srv->out)
		fclose(srv->out);

	free(srv->l);
	free(srv);
}

int
ircget(struct ircsrv *srv, struct ircmsg **rmsg)
{
	char		*cp, *tp;
	ssize_t		ls;

	if (!srv)
		return -1;

	if ((ls = getline(&srv->l, &srv->bs, srv->in)) == -1)
		return -1;

	if (ls == 0 || ls >= IRCMSGLEN)
		return -1;

	if (srv->l[ls - 2] != '\r' || srv->l[ls - 1] != '\n') {
		if (*rmsg)
			explicit_bzero(*rmsg, sizeof(struct ircmsg));
		return 0;
	}

	srv->l[ls - 2] = '\0';
	cp = srv->l;

	if ((tp = strsep(&cp, " ")) == NULL)
		return -1;

	if (!*rmsg) {
		*rmsg = newmsg();
		if (!*rmsg)
			return -1;
	} else {
		explicit_bzero(*rmsg, sizeof(struct ircmsg));
	}

	if (tp[0] == '@') {
		if (parsetags(tp, *rmsg) == -1)
			return -1;

		if ((tp = strsep(&cp, " ")) == NULL)
			return -1;
	}

	if (strlcpy((*rmsg)->cmd, tp, IRCCMDLEN) >= IRCCMDLEN)
		return -1;

	if (cp)
		if (strlcpy((*rmsg)->text, cp, IRCTXTLEN) >= IRCTXTLEN)
			return -1;

	return 0;
}

int
ircmsgcode(struct ircsrv *srv, enum irccode num, const char *nick, const char *fmt, ...)
{
	struct ircmsg	msg;
	va_list		ap;

	explicit_bzero(&msg, sizeof(struct ircmsg));

	if (snprintf(msg.src, IRCSRCLEN, ":%s", srv->host) >= IRCSRCLEN)
		return -1;

	if (snprintf(msg.cmd, IRCCMDLEN, "%03d %s", num,
	    nick) >= IRCCMDLEN)
		return -1;

	va_start(ap, fmt);

	if (vsnprintf(msg.text, IRCTXTLEN, fmt, ap) >= IRCTXTLEN)
		return -1;

	va_end(ap);

	return ircmsgsend(srv, &msg);
}

int
ircmsgjoin(struct ircsrv *srv, const char *nick, const char *chan)
{
	struct ircmsg	msg;

	explicit_bzero(&msg, sizeof(struct ircmsg));

	if (snprintf(msg.src, IRCSRCLEN, ":%s!~%s@%s", nick, nick, srv->host)
	    >= IRCSRCLEN)
		return -1;

	if (strlcpy(msg.cmd, "JOIN", IRCCMDLEN) >= IRCCMDLEN)
		return -1;

	if (strlcpy(msg.text, chan, IRCTXTLEN) >= IRCTXTLEN)
		return -1;

	return ircmsgsend(srv, &msg);
}

int
ircmsgping(struct ircsrv *srv, const char *nick, const char *text)
{
	struct ircmsg	msg;

	explicit_bzero(&msg, sizeof(struct ircmsg));

	if (snprintf(msg.src, IRCSRCLEN, ":%s!~%s@%s", nick, nick, srv->host)
	    >= IRCSRCLEN)
		return -1;

	if (strlcpy(msg.cmd, "PING", IRCCMDLEN) >= IRCCMDLEN)
		return -1;

	if (snprintf(msg.text, IRCTXTLEN, "%s", text) >= IRCTXTLEN)
		return -1;

	return ircmsgsend(srv, &msg);
}

int
ircmsgpong(struct ircsrv *srv, const char *nick, const char *text)
{
	struct ircmsg	msg;

	explicit_bzero(&msg, sizeof(struct ircmsg));

	if (snprintf(msg.src, IRCSRCLEN, ":%s!~%s@%s", nick, nick, srv->host)
	    >= IRCSRCLEN)
		return -1;

	if (strlcpy(msg.cmd, "PONG", IRCCMDLEN) >= IRCCMDLEN)
		return -1;

	if (snprintf(msg.text, IRCTXTLEN, "%s", text) >= IRCTXTLEN)
		return -1;

	return ircmsgsend(srv, &msg);
}


int
ircmsgpriv(struct ircsrv *srv, const char *nick, const char *target, const char *text)
{
	struct ircmsg	msg;

	explicit_bzero(&msg, sizeof(struct ircmsg));

	if (snprintf(msg.src, IRCSRCLEN, ":%s!~%s@%s", nick, nick, srv->host)
	    >= IRCSRCLEN)
		return -1;

	if (strlcpy(msg.cmd, "PRIVMSG", IRCCMDLEN) >= IRCCMDLEN)
		return -1;

	if (snprintf(msg.text, IRCTXTLEN, "%s :%s", target, text)
	    >= IRCTXTLEN)
		return -1;

	return ircmsgsend(srv, &msg);
}

int
ircmsgset(struct ircmsg **omsg, const char *src, const char *cmd, const char *fmt, ...)
{
	va_list		ap;

	if (!src || !cmd || !fmt)
		return -1;

	if (!*omsg) {
		*omsg = newmsg();
		if (!*omsg)
			return -1;
	}

	if (strlcpy((*omsg)->src, src, IRCSRCLEN) >= IRCSRCLEN)
		return -1;
	if (strlcpy((*omsg)->cmd, cmd, IRCCMDLEN) >= IRCCMDLEN)
		return -1;

	va_start(ap, fmt);

	if (vsnprintf((*omsg)->text, IRCTXTLEN, fmt, ap) >= IRCTXTLEN)
		return -1;

	va_end(ap);

	return 0;
}

char *
ircmsgcmd(struct ircmsg *msg)
{
	return msg->cmd;
}

int
ircmsgaddtag(struct ircmsg *msg, const char *key, const char *val)
{
	struct ircmsgtag	*t;

	if (!msg || !key)
		return -1;

	t = calloc(1, sizeof(struct ircmsgtag));
	if (!t)
		return -1;

	if (strlcpy(t->key, key, IRCTAGKEYLEN) >= IRCTAGKEYLEN) {
		free(t);
		return -1;
	}

	if (val)
		if (strlcpy(t->val, val, IRCTAGVALLEN) >= IRCTAGVALLEN) {
			free(t);
			return -1;
		}

	STAILQ_INSERT_TAIL(&(msg->tags), t, next);

	return 0;
}

void
ircmsgfree(struct ircmsg *msg)
{
	struct ircmsgtag	*tp, *ttp;

	if (!msg)
		return;

	STAILQ_FOREACH_SAFE(tp, &(msg->tags), next, ttp) {
		STAILQ_REMOVE(&(msg->tags), tp, ircmsgtag, next);
		free(tp);
	}

	free(msg);
}

int
ircmsgsend(struct ircsrv *srv, struct ircmsg *msg)
{
	struct ircmsgtag	*t, *l;

	if (!STAILQ_EMPTY(&(msg->tags))) {
		fputc('@', srv->out);

		l = STAILQ_LAST(&(msg->tags), ircmsgtag, next);

		STAILQ_FOREACH(t, &(msg->tags), next) {
			if (strlen(t->val) > 0)
				fprintf(srv->out, "%s=%s", t->key, t->val);
			else
				fprintf(srv->out, "%s", t->key);

			if (t != l)
				fputc(';', srv->out);
		}

		fputc(' ', srv->out);
	}

	fprintf(srv->out, "%s %s %s\r\n", msg->src, msg->cmd, msg->text);

	if (fflush(srv->out) == EOF)
		return -1;

	return 0;
}

char *
ircmsgsrc(struct ircmsg *msg)
{
	return msg->src;
}

char *
ircmsgtext(struct ircmsg *msg)
{
	return msg->text;
}

static struct ircmsg *
newmsg(void)
{
	struct ircmsg	*msg;

	msg = calloc(1, sizeof(struct ircmsg));
	if (!msg)
		return NULL;

	return msg;
}

static int
parsetags(char *line, struct ircmsg *msg)
{
	struct ircmsgtag	*tag;
	char		*tp, *ttp;

	STAILQ_INIT(&(msg->tags));

	while ((tp = strsep(&line, ";")) != NULL) {
		if (*tp == '\0')
			continue;

		tag = malloc(sizeof(struct ircmsgtag));
		if (!tag)
			return -1;

		if ((ttp = strsep(&tp, "=")) != NULL) {
			if (strlcpy(tag->key, ttp, IRCTAGKEYLEN) >= IRCTAGKEYLEN) {
				free(tag);
				return -1;
			}
			if (strlcpy(tag->val, tp, IRCTAGVALLEN) >= IRCTAGVALLEN) {
				free(tag);
				return -1;
			}
		} else {
			if (strlcpy(tag->key, tp, IRCTAGKEYLEN) >= IRCTAGKEYLEN) {
				free(tag);
				return -1;
			}
			tag->val[0] = '\0';
		}

		STAILQ_INSERT_TAIL(&(msg->tags), tag, next);
	}

	return 0;
}
