/*
 * Copyright (c) 2024 Andrea Biscuola <a@abiscuola.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/wait.h>

#include <err.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>
#include <api.h>

#include "compat.h"
#include "fchat.h"
#include "irc.h"

#define DEFCHAN	"#chat"
#define DEFHIST	-1
#define USERSMAX	2048

static char	*users[USERSMAX];
static const char *chan = DEFCHAN;
static const char *nick;
static char *url;
static int64_t id;
static pid_t repopid;

static __dead void	repopoller(struct ircsrv *, struct fchat *);
static __dead void	clientpoller(struct ircsrv *, struct fchat *);
static int		joinuser(struct ircsrv *, const char *);
static int		sendnames(struct ircsrv *);
static void		adduser(const char *);
static int		userexists(const char *);
static void		sighandler(int);
static __dead void	usage(void);

int
main(int argc, char *argv[])
{
	struct fchatmsgs	*fmsgs;
	struct fchatmsg	*fmsg;
	struct ircmsg	*imsg;
	struct fchat	*rchat;
	struct ircsrv	*srv;
	const char		*pass;
	int		c;
	int64_t		hist;

	chan = DEFCHAN;
	hist = DEFHIST;

	if (pledge("dns stdio inet proc rpath unveil", NULL) == -1)
		err(1, "pledge");

	if (unveil("/etc/ssl", "r") == -1)
		err(1, "unveil");
	if (unveil(NULL, NULL) == -1)
		err(1, "unveil");

	while ((c = getopt(argc, argv, "H:c:")) != -1)
		switch (c) {
		case 'H':
			hist = strtonum(optarg, 0, INT64_MAX, NULL);
			if (hist == 0)
				errx(1, "%s", "bad history value");
			hist = -hist;
			if (hist > 0)
				errx(1, "%s", "history value must be a positive number");
			break;
		case 'c':
			chan = optarg;
			if (chan[0] != '#')
				errx(1, "%s", "the channel name must start with '#'");
			break;
		case '?':
			usage();
		}

	argc -= optind;
	argv += optind;

	if (argc == 0)
		errx(1, "%s", "missing fossil chat url");

	/* Remove the trailing slash if present. Fossil does not like it */
	url = argv[0];
	if (url[strlen(url) - 1] == '/')
		url[strlen(url) - 1] = '\0';

	srv = ircnew(stdin, stdout);
	if (!srv)
		errx(1, "%s", "error while allocating irc server");

	nick = pass = NULL;
	imsg = NULL;

	/*
	 * Wait for PASS and NICK from the client.
	 *
	 * The information are used to login into the fossil repository
	 * chat.
	 */
	while (ircget(srv, &imsg) == 0) {
		if (!imsg)
			continue;

		if (strcmp(ircmsgcmd(imsg), "PASS") == 0) {
			pass = strdup(ircmsgtext(imsg));
			if (!pass) {
				ircmsgcode(srv, IRCERRPASSWDMISMATCH,
				    nick, ":Empty password");
				exit(1);
			}

			/* Skip the first column if it is there */
			if (pass[0] == ':')
				pass++;
		} else if (strcmp(ircmsgcmd(imsg), "NICK") == 0) {
			nick = strdup(ircmsgtext(imsg));
			if (!nick) {
				ircmsgcode(srv, IRCERRNONICKNAMEGIVEN,
				    nick, ":No nickname given");
				exit(1);
			}
		}

		if (nick && pass)
			break;
	}

	ircmsgfree(imsg);

	if (ircmsgcode(srv, IRCRPLWELCOME, nick, ":Welcome to the %s chat, %s",
	    url, nick) == -1)
		errx(1, "%s", "error while sending welcome IRC message");

	if (ircmsgcode(srv, IRCRPLWELCOME, nick, ":Trying to login with %s",
	    nick) == -1)
		errx(1, "%s", "error while sending welcome IRC message");

	/* Login to the fossil repository */
	rchat = fchatnew(url, nick, pass);
	if (!rchat)
		errx(1, "%s", "error while trying to login to the fossil repo chat");

	fmsgs = fchatpoll(rchat, hist);
	if (!fmsgs) {
		ircmsgcode(srv, IRCERRPASSWDMISMATCH, nick,
		    ":Wrong username or password");
		exit(1);
	}

	fmsg = NULL;
	if (fchatnext(fmsgs, &fmsg) == -1) {
		ircmsgcode(srv, IRCERRPASSWDMISMATCH, nick, ":%s", fchattext(fmsg));
		exit(1);
	}

	/*
	 * Here we check if the first polled message contains an error.
	 *
	 * Usually it means that the credentials provided are wrong and
	 * we will get a permission denied error message.
	 */
	if (fchaterror(fmsg)) {
		ircmsgcode(srv, IRCERRPASSWDMISMATCH, nick, ":%s", fchattext(fmsg));
		exit(1);
	}

	if (ircmsgcode(srv, IRCRPLLOGGEDIN, nick, ":You are now logged in as %s", nick)
	    == -1)
		errx(1, "error while sending confirmation of login");

	/*
	 * Login ok: welcome the client and let it join the channel
	 */

	adduser(nick);
	joinuser(srv, nick);

	/*
	 * If we fetched some history, send the messages to the user,
	 * otherwise, only save the message id for the following polling.
	 *
	 * While there, also send all the users we see in the messages to
	 * be listed by the client.
	 */
	do {
		id = fchatid(fmsg);

		if (hist == -1)
			break;

		if (fchatisdel(fmsg))
			continue;

		if (!userexists(fchatuser(fmsg))) {
			adduser(fchatuser(fmsg));
			joinuser(srv, fchatuser(fmsg));
		}

		if (ircmsgpriv(srv, fchatuser(fmsg), chan, fchattext(fmsg)) == -1)
			errx(1, "%s", "error while sending messages to the client");
	} while (fchatnext(fmsgs, &fmsg) == 0);

	sendnames(srv);

	fchatmsgsfree(fmsgs);

	/*
	 * Run the client poller and repository poller.
	 */
	switch ((repopid = fork())) {
	case 0:
		setproctitle("%s", "repopoller");
		repopoller(srv, rchat);
	default:
		/* We exit if the repopoller dies */
		signal(SIGCHLD, sighandler);

		setproctitle("%s", "clientpoller");
		clientpoller(srv, rchat);
	}

	return 0;
}

/*
 * clientpoller: fetch messages from the IRC client and forward them to
 * the fossil repository chat.
 *
 * We handle only PRIVMSG and PING here. Any other command receives
 * and "unknown command" reply.
 */
static __dead void
clientpoller(struct ircsrv *srv, struct fchat *repo)
{
	struct ircmsg	*msg;
	char		*txt;

	if (pledge("dns stdio inet rpath proc", NULL) == -1) {
		errx(1, "pledge");
	}

	msg = NULL;

	while (ircget(srv, &msg) == 0) {
		if (!msg)
			continue;

		if (strcasecmp(ircmsgcmd(msg), "PRIVMSG") == 0) {
			txt = ircmsgtext(msg);

			/*
			 * find the PRIVMSG arguments and remove the
			 * first colon character.
			 */

			if (!strsep(&txt, " "))
				continue;
			txt++;

			if (fchatsend(repo, txt, NULL) == -1)
				errx(1, "%s", "error sending to the repository");
		} else if (strcasecmp(ircmsgcmd(msg), "PING") == 0) {
			if (ircmsgpong(srv, ircmsgsrc(msg), ircmsgtext(msg)) == -1)
				errx(1, "%s", "error sending pong to the client");
		} else {
			if (ircmsgcode(srv, IRCERRUNKNOWNCOMMAND,
			    nick, "%s: unknown command", ircmsgcmd(msg))
			    == -1) {
				errx(1, "%s", "error sending to client");
			}
		}
	}

	kill(repopid, SIGTERM);
	waitpid(repopid, NULL, 0);

	exit(0);
}

/*
 * repopoller: fetch messages from the repository chat, and send those
 * to the IRC client.
 */
static __dead void
repopoller(struct ircsrv *srv, struct fchat *repo)
{
	struct fchatmsgs	*msgs;
	struct fchatmsg	*msg;

	if (pledge("dns stdio inet rpath", NULL) == -1) {
		warn("pledge");
		_exit(1);
	}

	msg = NULL;

	while ((msgs = fchatpoll(repo, id)) != NULL) {
		while (fchatnext(msgs, &msg) == 0) {
			if (fchaterror(msg)) {
				warnx("%s", fchattext(msg));
				_exit(1);
			}

			id = fchatid(msg);

			/*
			 * We do not want to forward messages related
			 * to "delete" actions, or messages coming from
			 * our own user.
			 */
			if (fchatisdel(msg) || strcmp(fchatuser(msg), nick) == 0)
				continue;

			/*
			 * If the user that sent the message does not exists,
			 * let the client know that it has "joined" the channel.
			 */
			if (!userexists(fchatuser(msg))) {
				adduser(fchatuser(msg));
				joinuser(srv, fchatuser(msg));
			}

			if (ircmsgpriv(srv, fchatuser(msg), chan, fchattext(msg)) == -1) {
				warnx("%s", "can not send to the irc client");
				_exit(1);
			}
		}

		fchatmsgsfree(msgs);
	}

	_exit(0);
}

static int
joinuser(struct ircsrv *srv, const char *user)
{
	if (ircmsgjoin(srv, user, chan) == -1)
		return -1;

	if (strcmp(user, nick) == 0) {
		if (ircmsgcode(srv, IRCRPLTOPIC, nick, "%s :The %s private chat", chan,
		    url) == -1)
			return -1;
	}

	return 0;
}

static int
sendnames(struct ircsrv *srv)
{
	char	**np;

	for (np = users; *np != NULL; np++)
		if (ircmsgcode(srv, IRCRPLNAMREPLY, nick, "= %s :%s", chan, *np) == -1)
			return -1;

	if (ircmsgcode(srv, IRCRPLENDOFNAMES, nick, "%s :End of /NAMES list", chan) == -1)
		return -1;

	return 0;
}

static void
adduser(const char *user)
{
	static char	**up = NULL;

	if (!up)
		up = users;

	if (up == users + USERSMAX - 1)
		return;

	*up = strdup(user);
	up++;
}

static int
userexists(const char *user)
{
	char **np;

	for (np = users; *np && np != users + USERSMAX - 1; np++)
		if (strcmp(user, *np) == 0)
			return 1;

	return 0;
}

void
sighandler(int sig)
{
	if (sig != SIGCHLD)
		return;

	waitpid(repopid, NULL, 0);
	exit(1);
}

static __dead void
usage(void)
{
	fprintf(stderr, "usage: %s [-H messages] [-c channel] url\n", getprogname());

	exit(1);
}
