/*	$Id: api.c,v 1.33 2021/03/11 10:12:01 absc Exp $	*/
/*
 * Copyright (c) 2018-2020 Andrea Biscuola <a@abiscuola.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <assert.h>
#include <paths.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <curl/curl.h>

#include "compat.h"
#include "jsmn.h"
#include "api.h"

#define BLEN	2048
#define STRLEN	1024

#define JTOKMIN	32

struct request {
	reqtype	type;
	char	*key;
};

static struct request requests[] = {
	{APIREQGET, "GET"},
	{APIREQPOST, "POST"},
	{APIREQPUT, "PUT"},
	{APIREQDEL, "DELETE"},
	{APIREQERR, NULL}
};

static int		setreqtype(struct api *, const char *);
static int		growbuf(struct api *, size_t);
static size_t	rpcwrite(char *, size_t, size_t, void *);
static size_t	rpcread(char *, size_t, size_t, void *);
static int		jsoninit(struct api *);

int
apiglobalinit(char **error)
{
	CURLcode	e;

	if ((e = curl_global_init(CURL_GLOBAL_ALL)) != CURLE_OK) {
		if (!*error)
			*error = strdup(curl_easy_strerror(e));
		return -1;
	}

	return 0;
}

struct api *
apiinit(const char *rt, const char *url)
{
	struct api	*rpc;

	if (!rt || !url)
		return NULL;

	rpc = calloc(1, sizeof(struct api));
	if (!rpc)
		return NULL;

	rpc->b = malloc(BLEN * sizeof(char));
	if (!rpc->b) {
		apifree(rpc);
		return NULL;
	}
	rpc->np = rpc->b;
	rpc->bs = BLEN;

	rpc->hnd = curl_easy_init();
	if (!rpc->hnd) {
		apifree(rpc);
		return NULL;
	}

	if (setreqtype(rpc, rt) == -1) {
		apifree(rpc);
		return NULL;
	}

	if (strlcpy(rpc->url, url, APIURLLEN) >= APIURLLEN) {
		apifree(rpc);
		return NULL;
	}

	curl_easy_setopt(rpc->hnd, CURLOPT_WRITEDATA, rpc);
	curl_easy_setopt(rpc->hnd, CURLOPT_WRITEFUNCTION, rpcwrite);
	curl_easy_setopt(rpc->hnd, CURLOPT_READFUNCTION, rpcread);

	if (getenv("LIBAPI_VERBOSE"))
		curl_easy_setopt(rpc->hnd, CURLOPT_VERBOSE, 1L);

	return rpc;
}

void
apifree(struct api *rpc)
{
	if (!rpc)
		return;

	curl_mime_free(rpc->mime);
	curl_slist_free_all(rpc->hdr);
	curl_easy_cleanup(rpc->hnd);

	freezero(rpc->b, rpc->bs);
	free(rpc->tk);
	free(rpc);
}

int
apiaddformfield(struct api *rpc, const char *n, const char *v,
    const char *f, size_t s)
{
	curl_mimepart	*p;

	if (!rpc || !n || !v)
		return -1;

	if (!rpc->mime) {
		rpc->mime = curl_mime_init(rpc->hnd);
		if (!rpc->mime)
			return -1;
	}

	p = curl_mime_addpart(rpc->mime);
	if (!p)
		return -1;

	if (curl_mime_data(p, v, s > 0 ? s : CURL_ZERO_TERMINATED) != CURLE_OK)
		return -1;

	if (f)
		if (curl_mime_filename(p, f) != CURLE_OK)
			return -1;

	if (curl_mime_name(p, n) != CURLE_OK)
		return -1;

	return 0;
}

int
apiaddstr(struct api *rpc, const char *n, const char *v)
{
	size_t	l, s;

	/*
	 * We assume the string is properly terminated with '\0'.
	 * Be sure to terminate the array before passing it to the
	 * function.
	 */

	if (!rpc || !v)
		return -1;

	l = strlen(v);

	/*
	 * We need to have enough space for 4 more characters
	 * in the buffer out of the string size we calculated.
	 */
	if (growbuf(rpc, l + 4) == -1)
		return -1;

	rpc->np[0] = '"';
	rpc->np++;
	rpc->s++;

	if (n) {
		s = snprintf(rpc->np, rpc->bs - rpc->s, "%s\":\"", n);
		if (s >= rpc->bs - rpc->s)
			return -1;

		rpc->s += s;
		rpc->np += s;
	}

	if ((s = strlcpy(rpc->np, v, rpc->bs - rpc->s)) >= rpc->bs - rpc->s)
		return -1;

	rpc->np += s;
	rpc->np[0] = '"';
	rpc->np[1] = ',';
	rpc->np[2] = '\0';
	rpc->s += (s + 2);
	rpc->np += 2;

	return 0;
}

int
apiaddint(struct api *rpc, const char *n, const long long v)
{
	int	s;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	if (!n)
		s = snprintf(rpc->np, rpc->bs - rpc->s, "%lld,", v);
	else
		s = snprintf(rpc->np, rpc->bs - rpc->s, "\"%s\":%lld,", n, v);
	if (s >= STRLEN)
		return -1;

	rpc->s += s;
	rpc->np += s;

	return 0;
}

int
apiaddfloat(struct api *rpc, const char *n, const double v)
{
	int	s;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	if (!n)
		s = snprintf(rpc->np, rpc->bs - rpc->s, "%.3f,", v);
	else
		s = snprintf(rpc->np, rpc->bs - rpc->s, "\"%s\":%.3f,", n, v);
	if (s >= STRLEN)
		return -1;

	rpc->s += s;
	rpc->np += s;

	return 0;
}

int
apiaddbool(struct api *rpc, const char *n, const apibool v)
{
	int	s;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	if (!n)
		s = snprintf(rpc->np, rpc->bs - rpc->s, "%s,",
		    v == APIFALSE ? "false" : "true");
	else
		s = snprintf(rpc->np, rpc->bs - rpc->s, "\"%s\":%s,", n,
		    v == APIFALSE ? "false" : "true");
	if (s >= STRLEN)
		return -1;

	rpc->s += s;
	rpc->np += s;

	return 0;
}

int
apiaddobj(struct api *rpc, const char *n)
{
	int	s;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	if (n)
		s = snprintf(rpc->np, rpc->bs - rpc->s, "\"%s\":{", n);
	else {
		s = snprintf(rpc->np, rpc->bs - rpc->s, "%s", "{");
	}
	if (s >= STRLEN)
		return -1;

	rpc->s += s;
	rpc->np += s;

	return 0;
}

int
apicloseobj(struct api *rpc)
{
	char	*p;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	p = rpc->np;
	if (p[-1] == ',') {
		p--;
		rpc->s--;
	}

	p[0] = '}';
	p[1] = ',';
	p[2] = '\0';
	p += 2;
	rpc->np = p;
	rpc->s += 2;

	return 0;
}

int
apiaddarray(struct api *rpc, const char *n)
{
	int	s;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	if (n)
		s = snprintf(rpc->np, rpc->bs - rpc->s, "\"%s\":[", n);
	else
		s = snprintf(rpc->np, rpc->bs - rpc->s, "[");
	if (s >= STRLEN)
		return -1;

	rpc->s += s;
	rpc->np += s;

	return 0;
}

int
apiclosearray(struct api *rpc)
{
	char	*p;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	p = rpc->np;
	if (p[-1] == ',') {
		p--;
		rpc->s--;
	}

	p[0] = ']';
	p[1] = ',';
	p[2] = '\0';
	p += 2;
	rpc->s += 2;
	rpc->np = p;

	return 0;
}

int
apiaddnull(struct api *rpc, const char *n)
{
	int	s;

	if (!rpc || !n)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	s = snprintf(rpc->np, rpc->bs - rpc->s, "\"%s\":%s,", n, "null");
	if (s >= STRLEN)
		return -1;

	rpc->s += s;
	rpc->np += s;

	return 0;
}

int
apiclose(struct api *rpc)
{
	char	*p;

	if (!rpc)
		return -1;

	if (growbuf(rpc, STRLEN) == -1)
		return -1;

	p = rpc->np;
	if (p[-1] == ',') {
		p--;
		rpc->s--;
	}

	p[0] = rpc->b[0] == '{' ? '}' : ']';
	p[1] = '\0';
	p++;
	rpc->s++;
	rpc->np = p;

	return 0;
}

int
apiaddheader(struct api *rpc, const char *key, const char *val)
{
	char	head[STRLEN];

	if (!rpc || !key || !val)
		return -1;

	if (snprintf(head, STRLEN, "%s: %s", key, val) >= STRLEN)
		return -1;

	rpc->hdr = curl_slist_append(rpc->hdr, head);

	return 0;
}

int
apisetauth(struct api *rpc, const char *user, const char *password)
{
	if (!rpc || !user || !password)
		return -1;

	curl_easy_setopt(rpc->hnd, CURLOPT_HTTPAUTH, CURLAUTH_BASIC);
	curl_easy_setopt(rpc->hnd, CURLOPT_USERNAME, user);
	curl_easy_setopt(rpc->hnd, CURLOPT_PASSWORD, password);

	return 0;
}

int
apisetbearer(struct api *rpc, const char *token)
{
	if (!rpc || !token)
		return -1;

	curl_easy_setopt(rpc->hnd, CURLOPT_HTTPAUTH, CURLAUTH_BEARER);
	curl_easy_setopt(rpc->hnd, CURLOPT_XOAUTH2_BEARER, token);

	return 0;
}

int
apisetcapath(struct api *rpc, const char *p)
{
	if (!rpc || !p)
		return -1;

	curl_easy_setopt(rpc->hnd, CURLOPT_CAPATH, p);

	return 0;
}

int
apisetcookiefile(struct api *rpc, const char *p)
{
	if (!rpc)
		return -1;

	if (p) {
		if (strlcpy(rpc->cf, p, PATH_MAX) >= PATH_MAX)
			return -1;

		curl_easy_setopt(rpc->hnd, CURLOPT_COOKIEJAR, rpc->cf);
	} else {
		strcpy(rpc->cf, "");
	}

	curl_easy_setopt(rpc->hnd, CURLOPT_COOKIEFILE, rpc->cf);

	return 0;
}

char *
apicookiefile(struct api *rpc)
{
	if (!rpc)
		return NULL;

	return strdup(rpc->cf);
}

int
apido(struct api *rpc)
{
	CURLcode	e;

	if (!rpc)
		return -1;

	if (rpc->mime) {
		curl_easy_setopt(rpc->hnd, CURLOPT_MIMEPOST, rpc->mime);
	} else if (rpc->s > 0) {
		curl_easy_setopt(rpc->hnd, CURLOPT_POSTFIELDSIZE_LARGE, rpc->s);
		curl_easy_setopt(rpc->hnd, CURLOPT_READDATA, rpc);
		curl_easy_setopt(rpc->hnd, CURLOPT_INFILESIZE, rpc->s);
	}

	curl_easy_setopt(rpc->hnd, CURLOPT_HTTPHEADER, rpc->hdr);
	curl_easy_setopt(rpc->hnd, CURLOPT_URL, rpc->url);

	rpc->np = rpc->b;

	e = curl_easy_perform(rpc->hnd);
	switch (e) {
	case CURLE_OK:
		break;
	case CURLE_OPERATION_TIMEDOUT:
		return -2;
	default:
		strlcpy(rpc->b, curl_easy_strerror(e), rpc->bs);

		return -1;
	}

	return 0;
}

char *
apierror(struct api *rpc)
{
	if (!rpc)
		return NULL;

	return strdup(rpc->b);
}

int
apireset(struct api *rpc, const char *rt, const char *url)
{
	if (!rpc)
		return -1;

	if (url)
		if (strlcpy(rpc->url, url, APIURLLEN) >= APIURLLEN)
			return -1;

	if (rt)
		if (setreqtype(rpc, rt) == -1)
			return -1;

	curl_mime_free(rpc->mime);
	rpc->mime = NULL;

	rpc->np = rpc->b;
	rpc->s = 0;
	rpc->tc = 0;
	rpc->cp = 0;

	return 0;
}

long
apihttpcode(struct api *rpc)
{
	long	hs;
	CURLcode	e;

	if (!rpc)
		return -1;

	e = curl_easy_getinfo(rpc->hnd, CURLINFO_RESPONSE_CODE, &hs);
	if (e != CURLE_OK)
		return -1;

	return hs;
}

ssize_t
apidump(struct api *rpc, char **buf)
{
	if (!rpc)
		return -1;

	*buf = malloc(rpc->s * sizeof(char));
	if (!*buf)
		return -1;

	/* This is just a dump, so we don't zero-terminate it */
	memcpy(*buf, rpc->b, rpc->s);

	return rpc->s;
}

int
apijsonnext(struct api *rpc, struct apiv *v)
{
	jsmntok_t	*tp;
	int	s, e;

	if (!rpc || !v)
		return -1;

	if (rpc->tc == 0) {
		if (jsoninit(rpc) == -1)
			return -1;
		tp = rpc->tp;

	} else {
		if (rpc->cp == rpc->tc)
			return -1;
		tp = rpc->tp;
		tp++;
		rpc->cp++;
	}

	rpc->tp = tp;

	v->val = NULL;

	switch (tp->type) {
	case JSMN_PRIMITIVE:
		v->type = APIJSONVAL;
		break;
	case JSMN_OBJECT:
		v->type = APIJSONOBJ;
		v->size = tp->size;
		return 0;
	case JSMN_ARRAY:
		v->type = APIJSONARR;
		v->size = tp->size;
		return 0;
	case JSMN_STRING:
		v->type = APIJSONSTR;
		break;
	default:
		return -1;
	}

	s = tp->start;
	e = tp->end;
	if (v->type == APIJSONSTR || v->type == APIJSONVAL) {
		v->val = rpc->b + s;
		v->val[e - s] = '\0';
	}

	return 0;
}

static int
setreqtype(struct api *rpc, const char *rtype)
{
	struct request  *rt;

	for (rt = requests; rt->type != APIREQERR; rt++)
		if (strcmp(rtype, rt->key) == 0)
			break;

	switch (rt->type) {
	case APIREQPOST:
		if (curl_easy_setopt(rpc->hnd, CURLOPT_POST, 1L) != CURLE_OK)
			return -1;
		break;
	case APIREQPUT:
		if (curl_easy_setopt(rpc->hnd, CURLOPT_PUT, 1L) != CURLE_OK)
			return -1;
		break;
	case APIREQGET:
		if (curl_easy_setopt(rpc->hnd, CURLOPT_HTTPGET, 1L) != CURLE_OK)
			return -1;
		break;
	case APIREQDEL:
		if (curl_easy_setopt(rpc->hnd, CURLOPT_CUSTOMREQUEST,
		    rt->key) != CURLE_OK)
			return -1;
		break;
	default:
		return -1;
	}

	return 0;
}

static int
growbuf(struct api *rpc, size_t psize)
{
	char	*p;

	while (rpc->bs - rpc->s <= psize) {
		p = reallocarray(rpc->b, rpc->bs * 2, sizeof(char));
		if (!p)
			return -1;
		rpc->b = p;

		rpc->np = rpc->b + rpc->s;
		rpc->bs = rpc->bs * 2;
	}

	assert(rpc->s + psize < rpc->bs);

	return 0;
}

static size_t
rpcwrite(char *p, size_t s, size_t n, void *ud)
{
	struct api	*r;

	r = ud;
	if (growbuf(r, s * n) == -1)
		return 0;

	if (r->s == 0)
		r->np = r->b;

	memcpy(r->np, p, n * s);

	r->s += s * n;
	r->np += n;
	r->np[0] = '\0';

	return n * s;
}

static size_t
rpcread(char *p, size_t s, size_t n, void *ud)
{
	struct api	*r;
	size_t	sz;

	r = ud;
	sz = r->s < n * s ? r->s : n * s;

	memcpy(p, r->np, sz);

	r->s -= sz;
	r->np += sz;

	return sz;
}

static int
jsoninit(struct api *rpc)
{
	jsmn_parser	jp;
	jsmntok_t		*tk, *p;

	rpc->ts = rpc->ts == 0 ? JTOKMIN : rpc->ts;
	tk = rpc->tk;

	jsmn_init(&jp);

	do {
		p = reallocarray(tk, rpc->ts, sizeof(jsmntok_t));
		if (!p)
			return -1;
		tk = p;

		rpc->tc = jsmn_parse(&jp, rpc->b, rpc->s, tk, rpc->ts);
		rpc->ts *= rpc->tc == JSMN_ERROR_NOMEM ? 2 : 1;
	} while (rpc->tc == JSMN_ERROR_NOMEM);

	rpc->cp = 1;
	rpc->tk = tk;
	rpc->tp = rpc->tk;

	return rpc->tc;
}
