/*	$Id: simple-get.c,v 1.3 2020/06/20 22:45:16 absc Exp $	*/
/*
 * Copyright (c) 2020 Andrea Biscuola <a@abiscuola.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * simple-get: basic GET example, fetching a JSON body from a postman
 * endpoint.
 *
 * This example, pass an argument to the endpoint as an encoded
 * URL one. The remote endpoint will give back a JSON containing
 * all the call details, including the argument itself.
 *
 * If the call is ok, we loop over the returned JSON searching for
 * our argument and printing it on standard ouput once found.
 *
 * A curl version of this call is:
 *
 *	curl -XGET 'https://postman-echo.com/get?api=api1'
 */

#include <err.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <curl/curl.h>
#include <api.h>

int
main(int argc, char *argv[])
{
	struct apiv	v;
	struct api		*rpc;
	char		*e;

	/* Initialize the library */
	if (apiglobalinit(&e) == -1)
		errx(1, "%s", e);

	/*
	 * Structure initialization. We want to do a GET
	 * to the postman echo service.
	 *
	 * We pass an "api" argument in the URL and we will
	 * read it back in the response.
	 */
	rpc = apiinit("GET", "https://postman-echo.com/get?api=api1");
	if (!rpc)
		errx(1, "%s", "initialization failed");

	/* We don't pass any body, just do the request */
	if (apido(rpc) == -1)
		errx(1, "%ld: %s", apihttpcode(rpc), apierror(rpc));

	/*
	 * Loop over the JSON response, we skip any token
	 * that's not an APIJSONSTR.
	 *
	 * When we find the "api" argument, we fetch the
	 * next token (the value) and print it on
	 * standard output.
	 */
	while (apijsonnext(rpc, &v) != -1) {
		if (v.type != APIJSONSTR)
			continue;

		if (strcmp(v.val, "api") == 0) {
			if (apijsonnext(rpc, &v) == -1)
				return 1;
			printf("%s: %s\n", "api", v.val);
		}
	}

	return 0;
}
