# Writing an IRCTk extension

## 1. Introduction

Extensions are the mechanism used to extend IRCTk capabilities. From simple
functions like searching in the conversations logs, to interfaces with third
party applications (think a SIP client, for example).

This document is aimed to anybody with programming experience that can
handle the concepts of reading and writing to standard input, standard output
and standard error, plus, some additional error handling. The target audience
for this guide, is anybody that wants to implement extensions from scratch,
without relying on an existing module.

An officially supported module for TCL is included with the IRCTk source tree,
taking care of all the repetitive parts for you.

## 2. Programming model

IRCTk extension architecture is unusual compared to other IRC clients. Where
tools like Hexchat directly embeds runtimes like perl, exposing to the plugin
complex and advanced data structures, IRCTk runs extensions as totally
separate programs. The communication between an extension and the
application happens through a two-way pipe using the dedicated
[extensions protocol](/doc/trunk/www/wiki/extensions.md),
plus, a series of environment variables, passed by IRCTk to a launched
extension.

## <a id="extensionvariables"></a>3. Environment variables

IRCTk, upon starting an extension executable, binds a series of environment
variables, useful for an extension to get configuration parameters related to
the application.

The environment variables are:

* **IRCTK_TCL_SDK**: path to the official TCL sdk, Useful only if you are
    writing extensions in TCL.
* **IRCTK_LOGS_PATH**: main directory of the IRCTk chat logs.

Here a small schema of the extensions subsystem, with some extensions "running":

``` pikchr
down
EXTS:	box rad 5px "Extensions subsystem" fit
	move

EXT1:	box rad 5px "Joiner" "(extension process)" fit
EXT2:	box rad 5px with .e at 2cm left of EXT1.w "Search" "(extension process)" fit
EXT3:	box rad 5px with .w at 2cm right of EXT1.e "Plumb" "(extension process)" fit

	spline <-> from EXTS.w then right until even with EXT2.n \
	    then down to EXT2.n
	spline <-> from EXTS.e then left until even with EXT3.n \
	    then down to EXT3.n
	spline <-> from EXTS.s to EXT1.n
```

In the image above shows an extension where the IRCTk subsystem is running
three extensions as external processes. A "Joiner" extension a "Search" extension
and a "Plumb" extension.

Being extensions external processes, they can be written in any programming
language. In general, an extension must:

1. Be a long-living process.
2. Be able to properly read and write from stdin, stdout and stderr.

The main advantage of this programming model is that IRCTk will never be
blocked by a running extension and an extension will never be blocked by IRCTk
performing intensive tasks. Also, IRCTk and it's extensions will run in separate
address spaces, leaving to the operating system things like randomizing their
stacks or, if an extension is written with support for it, to take advantage of
sandboxing technologies like pledge(2) and unveil(2).

## 2.1 I/O

As mentioned earlier, an extension must be able to read messages from it's
standard input channel and write to it's standard output channel.

Given that IRCTk treats the communication channels as asynchronous, it's
up to the extension developer to choose the I/O model in it's code. In general,
given that extensions are separate processes, there is no need for the
programmer to go the way of non-blocking I/O, and the suggested way is
to, always, flush the standard output after every message posted there.

This helps keep the buffers utilization low and makes IRCTk more reactive
without waiting for a whole lot of batched messages to be flushed. However,
in the end, it doesn't really matter to IRCTk itself.

## 2.2 Error handling

A extension willing to report an error to IRCTk, should write it in it's standard
error channel, that IRCTk will pick-up once the extension process terminates,
if possible. However, given the asynchronous nature of how IRCTk handles
extension channels, it's not guaranteed that the message will be reported
by IRCTk.

Another, more effective way of reporting errors, is for an extension to send
a PRIVMSG to IRCTk in a dedicated read-only channel. This guarantees that
the error will be visualized by the application.

Of course, any other way of reporting diagnostic messages outside of IRCTk
is possible. Naturally doing so, will depend on the specific use-case.

## 2.3 Process management

By design, IRCTk does not manage the lifetime of the extensions processes.
It's responsibility of an extension itself to properly terminate when an
unexpected condition happen.

The way employed by IRCTk to signal an extension peocess to terminate, is
to close the communication channel with the extension itself and deallocate
all the related resources from the extensions subsystem. If an extension does
not properly handle this specific situation, by checking the EOF condition
on the stdin file descriptor, the extension process will remain up and
running, albeit totally unable to communicate with the application itself.

In case an extension process remains in a "zombie" state, it's safe to terminate
it using the [kill(1)](http://man.openbsd.org/kill.1) command.

## 3. Example

Let's write a basic "echo" extension, that simply echoes back a message
received with an ECHO command. We will use TCL in this example, but
as said, any programming language could be used. We will start from snippets,
top to bottom and, at the end, we will link to the complete example. In the
example description, some part of the code are elided, simply because those
are not important in this case

For first, we need to declare how the kernel is going to run the extension.
In our case, it will be through tclsh:

```
#!/usr/bin/env tclsh8.6
```

Set some details about the extension:

```
set name "Echo"
set version 0.1
set protoversion 1.0
```

Now the main loop starts. It's just a while. In Tcl, gets returns a length
less-than zero, if something wrong happened while trying to read
standard input, for example, when IRCTk closes it's end of the pipe:

```
#
# Just cycle through stdin until the extension is closed.
#
while {[gets stdin line] >= 0} {
```

Split the message line, as per the protocol, fields are separated by a
tabulation character:

```
#
# Remember that the messages are exchanged as UTF-8
#
set msg [split [encoding convertfrom utf-8 $line] "\t"]
```

We just use a switch to check the message type (as per the protocol definition
it's the second field):

```
switch -exact -- [lindex $msg 1] {
```

First, we check if we received an handshake message. IRCTk sends this at the
beginning, right after starting the extension execution. If the protocol check
looks good, what happen next is:

1. The extension acknowledge the IRCTk handshake.
2. Run the extension's bit of the handshake.
3. Ask IRCTk to send the extension only commands of type irc.
4. Ask IRCTk to send only "echo" commands.

```
handshake {
	set id [lindex $msg 0]
	#
	# Run the handshake. The protocol is retro-compatible,
	# so we check if the version in our extension is too new
	# for the version of IRCTk we are running.
	#
	if {[lindex $msg 2] < $::protoversion} {
		nack $id "$::name: Incompatible protocol version"
		exit 1
	} else {
		ack $id
		handshake
		filter irc
		filter echo
		flush stdout
	}
}
```

Of course, there is no such a thing as an "ECHO" command in the IRC
protocol. However, IRCTk does not distinguish between IRC commands and
extended commands provided by an extension. The application will just
forward commands with their content, when the double forward-slash
mechanism ("//") is used by the user.

In the same switch, we handle the IRC commands. Given that we only receive
"echo", it's easy to do:

```
irc {
	writemsg [lindex $msg 8] [lindex $msg 9] \
	    PRIVMSG [join [lrange $msg 12 end]]
	flush stdout
}
```

What the extension do, in this case, is echoing back, as a PRIVMSG, the content
of an ECHO command received from the application. The PRIVMSG will be
sent to the (public or private) channel where the user sent the original
ECHO command.

That's it! The full code for the extension is in the [echo.tcl](/uv/echo-0.1.tcl) file.
